#!/usr/bin/python3
# coding=utf-8
##################################################################
#              ____     _     _ __  __                 
#             / __/__  (_)___(_) /_/ /  ___  ___  ___ _
#            _\ \/ _ \/ / __/ / __/ /__/ _ \/ _ \/ _ `/
#           /___/ .__/_/_/ /_/\__/____/\___/_//_/\_, / 
#              /_/                              /___/  
# 
# Copyright (c) 2025 Chongqing Spiritlong Technology Co., Ltd.
# All rights reserved.
# @author	arthuryang
# @brief	测试POST API的工具
##################################################################

import	json
import	requests

## 测试POST API，显示返回的数据
#	URL		协议+主机+端口
#	route		路由
#	post_json	要传递的json
def postman(URL, route, post_json, print_return=False):
	response	= requests.post(URL+route, json=post_json)
	print(f"\npost test: {URL+route:64}{response.status_code}")

	if print_return:
		try:
			j	= json.loads(response.text)
			for t in j:
				print(f"{t:32}{j[t]}")
		except Exception as e:
			print(e)

	return response.text


# 脚本执行
if __name__ == '__main__':
	postman("http://localhost:8080", "/test_DTO", {
		'group_ID'	: 1234,
		'fghCDEF'	: 'test2',
		'nobody'	: 'test3'
	}, True)

	



content	= '''#!/usr/bin/python3
# coding=utf-8
##################################################################
#               ____     _     _ __  __                 
#              / __/__  (_)___(_) /_/ /  ___  ___  ___ _
#             _\\ \\/ _ \\/ / __/ / __/ /__/ _ \\/ _ \\/ _ `/
#            /___/ .__/_/_/ /_/\\__/____/\\___/_//_/\\_, / 
#               /_/                              /___/  
# 
# Copyright (c) 2025 Chongqing Spiritlong Technology Co., Ltd.
# All rights reserved.
# @author	arthuryang
# @brief	webframe框架python后端，包括路由、uWSGI入口
##################################################################

import	sys
import	os
import	datetime
import	time
import	textwrap
import	redis
import	SpiritLong_utility as utility
import	SpiritLong_database as database
import	SpiritLong_excel as excel

from	routes		import routes

# 全局redis客户端
redis_client		= redis.Redis(host='localhost', port=6379, db=0, decode_responses=True)
# 全局配置
application_config	= {}
# 视图表配置
view_table_config	= {}

############################## 工具 ############################## 

## 更新全局配置
def update_application_config(config_file=None, config_db=None):
	global application_config

	file_path	= os.path.abspath(os.path.dirname(__file__))

	if config_file is None:
		# 默认的配置文件在./config/application.config
		config_file	= f"{file_path}/config/application.config"

	if not os.path.exists(config_file):
		# 无配置文件
		print(f"failed to update application_config for config_file not exists: {config_file}")
		return 
	# 读取配置文件
	with open(config_file, 'r') as f:
		for line in f:
			if line.strip().startswith('#'):
				continue
			parts	= line.split('=')
			if len(parts)==2:
				application_config[parts[0].strip()]	= parts[1].strip()
	# 检查主机是否为DEVELOP
	application_config['hostname']	= os.uname().nodename
	is_develop_host	= True if application_config['hostname'].startswith('DEVELOP') else False

	# 尝试读取配置数据库
	try:
		EXON_db	= database.Database({
			"host"		: application_config['EXON_application_host'],
			"user"		: application_config['EXON_application_user'],
			"passwd"	: application_config['EXON_application_password'],
			"database"	: application_config['EXON_application_database'],
			"port"		: int(application_config['EXON_application_port']),
		})


		result	= EXON_db.query_single(f\'\'\'
			SELECT 
				`APPLICATION`.`ID`,
				`APPLICATION`.`CODE`,
				`APPLICATION`.`NAME`,
				`APPLICATION`.`HTTP_PORT`,
				`APPLICATION`.`WEBSOCKET_PORT`,
				`APPLICATION`.`DEVELOP_HTTP_PORT`,
				`APPLICATION`.`DEVELOP_WEBSOCKET_PORT`,
				`APPLICATION`.`DEVELOP_REDIS_INDEX`,
				`DOMAIN`.`LEVEL2`,
				`DOMAIN`.`LEVEL1`,
				`PROJECT`.`CODE` AS `PROJECT_CODE`
			FROM `APPLICATION` 
			JOIN `DOMAIN`	ON `APPLICATION`.`DOMAIN_ID`=`DOMAIN`.`ID`
			JOIN `HOST`	ON `HOST`.`ID`=`DOMAIN`.`HOST_ID`
			JOIN `PROJECT`	ON `APPLICATION`.`PROJECT_ID`=`PROJECT`.`ID`
			WHERE `APPLICATION`.`ID`={application_config['EXON_application_ID']}
		\'\'\')
		application_config['application_code']	= result['CODE']
		application_config['application_name']	= result['NAME']
		application_config['http_port']		= result['DEVELOP_HTTP_PORT'] if is_develop_host else result['HTTP_PORT']
		application_config['websocket_port']	= result['DEVELOP_WEBSOCKET_PORT'] if is_develop_host else result['WEBSOCKET_PORT']
		application_config['redis_index']	= result['DEVELOP_REDIS_INDEX']	if is_develop_host else 0
		application_config['domain']		= f"{result['LEVEL2']}.{result['LEVEL1']}"
		application_config['project_code']	= result['PROJECT_CODE']
	except:
		print("failed to query EXON database. use default value")
		application_config['application_code']	= file_path.replace("/", "_").replace(".","")
		application_config['application_name']	= "未知项目"
		application_config['http_port']		= 8080
		application_config['websocket_port']	= 9000
		application_config['redis_index']	= 0
		application_config['domain']		= "unkownn_domain"
		application_config['project_code']	= "UNKNOWN"
	
	# 源代码路径
	application_config['path_source']		= file_path
	application_config['path_config']		= f"{application_config['path_source']}/config"
	application_config['path_frontend']		= f"{application_config['path_source']}/frontend"
	# 部署路径/链接
	application_config['path_base']			= f"/data/Application/{application_config['application_code']}"
	application_config['link_backend']		= f"{application_config['path_base']}/backend"
	application_config['path_frontend_runtime']	= f"{application_config['path_base']}/runtime/frontend"
	application_config['path_file']			= f"{application_config['path_base']}/file"
	application_config['path_log']			= f"{application_config['path_base']}/log"
	application_config['path_runtime']		= f"{application_config['path_base']}/runtime"
	application_config['link_uWSGI_logrotate']	= f"/etc/logrotate.d/uWSGI.{application_config['application_code']}"
	# uWSGI相关文件
	application_config['template_uWSGI_config']	= f"{application_config['path_config']}/uWSGI.config"
	application_config['template_uWSGI_logrotate']	= f"{application_config['path_config']}/logrotate.uWSGI.config"
	application_config['file_uWSGI_config']		= f"{application_config['path_runtime']}/uWSGI.config"
	application_config['file_uWSGI_socket']		= f"{application_config['path_runtime']}/uWSGI.socket"
	application_config['file_uWSGI_PID']		= f"{application_config['path_runtime']}/uWSGI.PID"
	application_config['file_uWSGI_reload']		= f"{application_config['path_runtime']}/uWSGI.reload"
	application_config['file_uWSGI_log_reopen']	= f"{application_config['path_runtime']}/uWSGI.log_reopen"
	application_config['file_uWSGI_logrotate']	= f"{application_config['path_runtime']}/uWSGI.logrotate"
	application_config['file_uWSGI_log']		= f"{application_config['path_log']}/uWSGI.log"
	# nginx相关文件
	application_config['file_certificate']		= f"{application_config['path_config']}/{application_config['domain']}.pem"
	application_config['file_certificate_key']	= f"{application_config['path_config']}/{application_config['domain']}.key"
	application_config['template_nginx_config']	= f"{application_config['path_config']}/nginx.config"
	application_config['file_nginx_config']		= f"{application_config['path_runtime']}/nginx.config"
	application_config['link_nginx_config']		= f"/etc/nginx/sites-enabled/{application_config['application_code']}"
	
	def set_default(name, default_value):
		global application_config
		if name not in application_config:
			application_config[name]	= default_value

	set_default('uWSGI_mount',		'/API') 
	set_default('uWSGI_process_n',		4) 
	set_default('uWSGI_request_max',	5000)
	set_default('uWSGI_buffer_size',	65536)
	set_default('uWSGI_post_buffering',	65536)

	# 数据库
	application_config['db']	= database.Database({
		'host'		: application_config['database_host'],
		'user'		: application_config['database_user'],
		'passwd'	: application_config['database_password'],
		'database'	: application_config['database_database'],
		'port'		: int(application_config['database_port']),
	}, use_redis={ 
		'host'	: application_config['redis_host'],
		'port'	: application_config['redis_port'],
		'db'	: application_config['redis_index'],
	})
	application_config['redis']	= application_config['db'].redis

# 加载视图表配置
def load_view_table_config():
	global view_table_config
	
	# 从config/view_table_config.xlsx中读取视图表配置
	config_file	= f"{application_config['path_config']}/view_table_config.xlsx"
	if not os.path.exists(config_file):
		print(f"未找到视图表配置文件: {config_file}")
		return
	try:
		book	= excel.open_xlsx(config_file)
		meta_sheet	= book['META']
		field_sheet	= book['FIELD']
		view_sheet	= book['VIEW']
		element_sheet	= book['ELEMENT']

		view_config		= excel.get_records_with_title(view_sheet)
		field_config		= excel.get_records_with_title(field_sheet)
		meta_config		= excel.get_records_with_title(meta_sheet)
		view_config_dict	= {}
		# 一个VIEW_NAME对应一个VIEW_TABLE_ID数组
		for view in view_config:
			if view['VIEW_NAME'] not in view_config_dict:
				view_config_dict[view['VIEW_NAME']]	= []
			view_config_dict[view['VIEW_NAME']].append(view['VIEW_TABLE_ID'])
		# 一个VIEW_TABLE_ID对应一个FIELD数组
		field_config_dict	= {}
		for field in field_config:
			if field['VIEW_TABLE_ID'] not in field_config_dict:
				field_config_dict[field['VIEW_TABLE_ID']]	= []
			field_config_dict[field['VIEW_TABLE_ID']].append(field['NAME'])
		# 一个VIEW_TABLE_ID对应一个VIEW_TABLE配置
		view_table_config_dict	= {}
		for view_table in meta_config:
			if view_table['VIEW_TABLE_ID'] not in view_table_config_dict:
				view_table_config_dict[view_table['VIEW_TABLE_ID']]	= {}
			view_table_config_dict[view_table['VIEW_TABLE_ID']]	= view_table
		view_table_config	= {
			'meta'		: view_table_config_dict,
			'element'	: excel.get_records_with_title(element_sheet),
			'field'		: field_config_dict,
			'view'		: view_config_dict,
		}
		print(f'加载视图表配置成功：{view_table_config}')
	except Exception as e:
		sys.exit(f"加载视图表配置失败: {e}")

# 加载所有权限列表
def load_all_permission_list():
	try:
		config_file	= f"{application_config['path_config']}/permission_config.xlsx"
		if not os.path.exists(config_file):
			print(f"未找到权限配置文件: {config_file}")
			return
		book	= excel.open_xlsx(config_file)
		sheet	= book['Sheet1']
		permission_list	= excel.get_records_with_title(sheet)
		permission_list	= [{
			'key'	: permission['序号'],
			'value'	: permission['权限'],
		} for permission in permission_list]
		redis_client.set('ALL_PERMISSION_LIST', utility.object_to_json(permission_list))
	except Exception as e:
		sys.exit(f"加载权限配置失败: {e}")

# 立即加载配置
update_application_config()
load_view_table_config()
load_all_permission_list()

############################## uWSGI ##############################

# 初始化全局路由
routes.redis		= application_config['redis']
routes.db		= application_config['db']

# 注意uwsgi并非一个python包，被uwsgi服务调用时才会有这个内建的模块
try: 
	import	uwsgi
	import	application.route
except ImportError:
	pass

## uWSGI入口
def application(environment, start):
	# ------------------ 路由前处理 ------------------

	# ------------------ 路由解析 ------------------
	data	= routes.query(environment)

	# ------------------ 路由后处理 ------------------
	# 增加日志变量
	uwsgi.set_logvar('datetime', str(datetime.datetime.today()))
	uwsgi.set_logvar('user_ID', str(data['user_ID']))

	# 权限管理
	
	# 执行路由，返回响应
	return	routes.response(data, start)

#################################### run.sh #################################
## 启动
#	backend_only	是否只启动后端
def webframe_run_start(backend_only=True):
	global application_config

	# 准备目录/符号链接：每个应用在/data/Application下有一个以其应用代码为名的目录
	os.system(f"mkdir -p {application_config['path_file']}")
	os.system(f"mkdir -p {application_config['path_log']}")
	os.system(f"mkdir -p {application_config['path_runtime']}")
	os.system(f"mkdir -p {application_config['path_frontend_runtime']}")
	# 第一次添加之后不再覆盖，防止从此路径引用启动
	if not os.path.exists(application_config['link_backend']):
		os.system(f"ln -sf {application_config['path_source']} {application_config['link_backend']}")

	def touch_file(filename, mode="666"):
		os.system(f"touch {filename}; chmod {mode} {filename}")
	touch_file(application_config['file_uWSGI_reload'])
	touch_file(application_config['file_uWSGI_log_reopen'])

	# ----------------- uWSGI -------------------
	# uWSGI配置
	with open(application_config['template_uWSGI_config'], "r") as f:
		content	= f.read()
	content	= content.replace('$APPLICATION_CODE',	application_config['application_code'])
	content	= content.replace('$MOUNT',		application_config['uWSGI_mount'])
	content	= content.replace('$APP_FILE',		os.path.basename(__file__))
	content	= content.replace('$SOCKET',		application_config['file_uWSGI_socket'])
	content	= content.replace('$PROCESS_N',		application_config['uWSGI_process_n'])
	content	= content.replace('$PID_FILE',		application_config['file_uWSGI_PID'])
	content	= content.replace('$LOG_FILE',		application_config['file_uWSGI_log'])
	content	= content.replace('$RELOAD_FILE',	application_config['file_uWSGI_reload'])
	content	= content.replace('$LOG_REOPEN',	application_config['file_uWSGI_log_reopen'])
	content	= content.replace('$REQUEST_MAX',	application_config['uWSGI_request_max'])
	content	= content.replace('$HARAKIRI',		application_config['uWSGI_harakiri'])
	content	= content.replace('$BUFFER_SIZE',	application_config['uWSGI_buffer_size'])
	content	= content.replace('$POST_BUFFERING',	application_config['uWSGI_post_buffering'])
	content	= content.replace('$LOGREOPEN_FILE',	application_config['file_uWSGI_log_reopen'])

	with open(application_config['file_uWSGI_config'], "w") as f:
		f.write(content)

	# 确保日志文件
	os.system(f"touch {application_config['file_uWSGI_log']}")

	# 配置日志切片 uWSGI
	with open(application_config['template_uWSGI_logrotate'], 'r') as f:
		content	= f.read()
	content	= content.replace('$UWSGI_LOG_FILE',		application_config['file_uWSGI_log'])
	content	= content.replace('$UWSGI_LOG_REOPEN',		application_config['file_uWSGI_log_reopen'])
	with open(application_config['file_uWSGI_logrotate'], "w") as f:
		f.write(content)

	# 链接logrotate
	os.system(f"ln -snf {application_config['file_uWSGI_logrotate']} {application_config['link_uWSGI_logrotate']}")

	# 启动uWSGI
	os.system(f"uwsgi --ini {application_config['file_uWSGI_config']}")
	
	# ----------------- frontend ------------------
	# 处理前端源代码目录下的内容：
	if not backend_only:
		for frontend in os.listdir(application_config['path_frontend']):
			frontend_path	= os.path.join(application_config['path_frontend'], frontend)
			# 前端为目录则需要打包
			if os.path.isdir(frontend_path):
				vite_config_file	= os.path.join(frontend_path, 'vite.config.js')
				if not os.path.exists(vite_config_file):
					# 跳过非vite项目
					print(f"跳过非vite项目: {frontend_path}")
					continue
				with open(vite_config_file, "r") as f:
					content	= f.read()
				output	= os.path.join(application_config['path_frontend_runtime'], frontend)
				# 替换vite.config.js中的配置build和base
				replace_flag_start	= '// TO REPLACE START'
				replace_flag_end	= '// TO REPLACE END'
				to_replace_content	= content.split(replace_flag_start)[1]
				if to_replace_content:
					to_replace_content	= to_replace_content.split(replace_flag_end)[0]
				if to_replace_content:
					replace_content	= textwrap.dedent(f\'\'\'
						build: {{
							outDir: '{output}'
						}},
						base: '/{frontend}/'
					\'\'\')
					replace_content	= '\\n'.join(['\\t' + line if line else line for line in replace_content.split('\\n')])+'\\t'
					content	= content.replace(to_replace_content, replace_content)
				with open(vite_config_file, "w") as f:
					f.write(content)
				# 先清除之前的打包内容
				if os.path.exists(output):
					os.system(f"rm -rf {output}/*")
				# 安装依赖和打包
				os.system(f\'\'\'
					cd {frontend_path}
					npm install
					npm run build
				\'\'\')
				# 删除node_modules
				if os.path.exists(f"{frontend_path}/node_modules"):
					os.system(f"rm -rf {frontend_path}/node_modules")


	# 检查SSL证书
	if not (os.path.exists(application_config['file_certificate']) and os.path.exists(application_config['file_certificate_key'])):
		# 没有证书就自行生成
		os.system(f"openssl genrsa -out {application_config['file_certificate_key']} 2048")
		os.system(f\'\'\'openssl req -new -x509 -key {application_config['file_certificate_key']} -out {application_config['file_certificate']} \'\'\'+
			f\'\'\'-days 1095 -subj "/C=CN/ST=CQ/L=BB/O=SpiritLong/CN={application_config['domain']}"\'\'\')


	# nginx配置：变量替换，设置链接
	with open(application_config['template_nginx_config'], 'r') as f:
		content	= f.read()
	content	= content.replace('$PORT',			str(application_config['http_port']))
	content	= content.replace('$CERTIFICATE_FILE',		application_config['file_certificate'])
	content	= content.replace('$CERTIFICATE_KEY_FILE',	application_config['file_certificate_key'])
	content	= content.replace('$DOMAIN',			application_config['domain'])
	content	= content.replace('$FRONTEND_PATH',		application_config['path_frontend_runtime'])
	content	= content.replace('$UWSGI_MOUNT',		application_config['uWSGI_mount'])
	content	= content.replace('$UWSGI_SOCKET',		application_config['file_uWSGI_socket'])
	content	= content.replace('$WEBSOCKET_PORT',		str(application_config['websocket_port']))
	# 微信认证的MP_verify文件，如果config目录下有MP_verify开头的文件，则复制到runtime目录下
	for file in os.listdir(application_config['path_config']):
		if file.startswith('MP_verify'):
			os.system(f"cp {application_config['path_config']}/{file} {application_config['path_frontend_runtime']}/{file}")
	# 遍历前端源代码目录，子目录名作为访问路径
	for frontend in os.listdir(application_config['path_frontend']):
		frontend_path	= os.path.join(application_config['path_frontend'], frontend)
		if not os.path.isdir(frontend_path):
			continue
		replace_content	= textwrap.dedent(f\'\'\'
			location / {{
				root {application_config['path_frontend_runtime']};
				index index.html;
				try_files $uri $uri/ /index.html;
			}}
			
			location /{frontend} {{
				alias {application_config['path_frontend_runtime']}/{frontend};
				index index.html;
				try_files $uri $uri/ /{frontend}/index.html;
			}}
		\'\'\')
		replace_content	= '\\n'.join(['\\t' + line if line else line for line in replace_content.split('\\n')])
		replace_flag	= '# frontend'
		if replace_flag in content:
			content	= content.replace(replace_flag, replace_content)
	with open(application_config['file_nginx_config'], "w") as f:
		f.write(content)
	
	os.system(f"ln -snf {application_config['file_nginx_config']} {application_config['link_nginx_config']}")
	
	# 设置权限
	os.system(f"chown www-data {application_config['path_runtime']} -R")
	os.system(f"chgrp www-data {application_config['path_runtime']} -R")


	# 启动nginx
	if os.system("service nginx reload")!=0:
		print("failed")

## 停止
def webframe_run_stop():
	global application_config
	# 刷新redis
	redis_client.flushdb()
	# 停止 uWSGI
	os.system(f"uwsgi --stop {application_config['file_uWSGI_PID']}")

# ===================================== 视图表相关API =====================================
# 生成视图表的所有ID列表
#	view_name	视图名称
#	condition	条件
#	返回值		视图表ID列表规则字符串
def get_view_table_IDs(view_name, condition=None, controller={}):
	views_config		= view_table_config['view']
	meta_config		= view_table_config['meta']
	if view_name not in views_config:
		print(f"配置文件的VIEW中没有{view_name}视图")
		return {}

	view_table_IDs	= views_config[view_name]
	view_ID_info	= {}
	for view_table_ID in view_table_IDs:
		# meta表和element表单独处理
		if view_table_ID==0 or view_table_ID==1:
			continue
		# 调用controller中对应的函数获取到视图表的所有ID
		function_name	= f"VD_{view_table_ID}_get_IDs"
		if not hasattr(controller, function_name):
			print(f"controller中没有{view_name}的get_IDs函数{function_name}")
			continue
		function	= getattr(controller, function_name)
		parameters	= condition.get(str(view_table_ID))
		ID_list		= function(**parameters) if parameters else function()
		if not isinstance(ID_list, list):
			print(f"controller中函数get_{view_name}的返回值必须是列表")
			ID_list	= []
		# 该是视图表所有ID
		all_ID_string			= stringify_ID_list(ID_list)
		# 初始订阅ID
		initial_values			= meta_config[view_table_ID]['INITIAL_VALUES']
		initial_subscribe_ID_string	= initial_values if initial_values else ''
		# ;之前是所有ID，之后是初始订阅ID
		view_ID_info[view_table_ID]	= f"{all_ID_string};{initial_subscribe_ID_string}"
		# 调用controller中对应的函数生成ID对应的视图表数据存到redis中
		function_name	= f"VD_{view_table_ID}_get_data"
		if not hasattr(controller, function_name):
			print(f"controller中没有{view_name}的get_data函数{function_name}")
			continue
		function	= getattr(controller, function_name)
		parameters	= condition.get(str(view_table_ID))
		data		= function(**parameters) if parameters else function()
		if not isinstance(data, list):
			print(f"controller中函数{function_name}的返回值必须是列表")
			continue
		for data_item in data:
			for view_table_record_ID in data_item:
				record_name	= f"VD {view_table_ID} {view_table_record_ID}"
				redis_client.set(record_name, utility.object_to_json(data_item[view_table_record_ID]))
	return view_ID_info

# 更新订阅
#	view_name	视图名称，订阅视图中的哪些表
#	tables		存储需要订阅或者退订的表数据ID信息，是一个字典
#	websocket_code	websocket code
def update_subscription(view_name, tables, websocket_code):
	global view_table_config
	# 订阅视图表（修改此处将ID转换为字符串）
	view_table_IDs	= [str(ID) for ID in view_table_config['view'][view_name]]
	channels	= [f"VD 0 {ID}" for ID in view_table_IDs]
	message		= f"SUBSCRIBE {websocket_code} {','.join(channels)}"
	print(f"订阅消息：{message}")
	redis_client.publish("WEBSOCKET SERVER", message)
	
	# 订阅表单元素表（同样处理元素ID）
	element_IDs	= [str(element['ELEMENT_ID']) for element in view_table_config['element']]
	channels	= [f"VD 1 {element_ID}" for element_ID in element_IDs]
	message		= f"SUBSCRIBE {websocket_code} {','.join(channels)}"
	print(f"订阅消息：{message}")
	redis_client.publish("WEBSOCKET SERVER", message)


	# 订阅视图表记录
	for table_ID, ID_string in tables.items():
		# 按规则解析ID_string
		subscribe_ID_string, unsubscribe_ID_string = ID_string.split(';')
		subscribe_ID_list	= parse_ID_string(subscribe_ID_string)
		unsubscribe_ID_list	= parse_ID_string(unsubscribe_ID_string)
		redis_client.set(f"VD 0 {table_ID}", subscribe_ID_string)
		# 处理订阅
		if subscribe_ID_list:
			subscribe_channels	= [f"VD {table_ID} {ID}" for ID in subscribe_ID_list]
			channels		= ','.join(subscribe_channels)
			message			= f"SUBSCRIBE {websocket_code} {channels}"
			print(f"订阅消息：{message}")
			redis_client.publish("WEBSOCKET SERVER", message)

		# 处理取消订阅
		if unsubscribe_ID_list:
			unsubscribe_channels	= [f"VD {table_ID} {ID}" for ID in unsubscribe_ID_list]
			channels		= ','.join(unsubscribe_channels)
			message			= f"UNSUBSCRIBE {websocket_code} {channels}"
			print(f"取消订阅消息：{message}")
			redis_client.publish("WEBSOCKET SERVER", message)
	time.sleep(0.02)
	# 发布数据
	publish_data(tables)

# 发布数据
#	websocket_code	websocket code
#	channels	频道
def publish_data(tables):
	for table_ID, ID_string in tables.items():
		# 按规则解析ID_string
		subscribe_ID_string = ID_string.split(';')[0]
		subscribe_ID_list	= parse_ID_string(subscribe_ID_string)
		for record_ID in subscribe_ID_list:
			record_data	= redis_client.get(f"VD {table_ID} {record_ID}")
			if record_data:
				record_data	= utility.json_to_object(record_data)
			channel		= f"VD {table_ID} {record_ID}"
			message		= utility.object_to_json((channel, record_data))
			redis_client.publish(channel, message)

# 对表发布数据
#	table_ID	表ID
#	websocket_code	websocket code
#	channels	频道
def publish_table_data(table_ID):
	record_ID_list	= redis_client.keys(f"VD {table_ID} *")
	for record_ID in record_ID_list:
		record_data	= redis_client.get(f"VD {table_ID} {record_ID}")
		if record_data:
			record_data	= utility.json_to_object(record_data)
		channel		= f"VD {table_ID} {record_ID}"
		message		= utility.object_to_json((channel, record_data))
		redis_client.publish(channel, message)

# 按规则解析ID_string，ID_string由逗号隔开，如果包含-则表示范围
#   ID_string   ID字符串，格式如: "1,2,3-5"
#   返回值      一个ID列表
def parse_ID_string(ID_string):
	if not ID_string.strip():
		return []
	result	= []
	for part in ID_string.split(','):
		if '-' in part:
			start, end	= map(int, part.split('-'))
			result.extend(range(start, end + 1))
		else:
			result.append(int(part))
	return sorted(result)

# 按规则将ID列表转换为字符串
#   ID_list   ID列表
#   返回值     ID字符串，格式如: "1,2,3-5"
def stringify_ID_list(ID_list):
	if not ID_list:
		return ""
	ID_list	= sorted(set(ID_list))
	result	= []
	start	= ID_list[0]
	for i in range(1, len(ID_list)):
		if ID_list[i] != ID_list[i - 1] + 1:
			result.append(f"{start}-{ID_list[i - 1]}" if start != ID_list[i - 1] else str(start))
			start = ID_list[i]
	result.append(f"{start}-{ID_list[-1]}" if start != ID_list[-1] else str(start))
	return ",".join(result)

class Logger:
	def __init__(self, db_config=None, log_dir='logs'):
		self.db		= db_config if db_config else None
		self.log_dir	= log_dir

	def _write_file_log(self, level, module, message):
		try:
			log_file = os.path.join(self.log_dir, f"runtime.log")
			timestamp = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
			
			with open(log_file, 'w', encoding='utf-8') as f:
				log_entry = f"[{timestamp}] [{level}] [{module}] {message}\\n"
				f.write(log_entry)
		except Exception as e:
			print(f"写入日志文件时发生错误：{str(e)}")

	def _write_db_log(self, log_data):
		self.db.insert('LOG', [log_data])
	
	def _log(self, level, module, message):
		self._write_file_log(level, module, message)
		
		log_data = {
			'TYPE'		: 'SYSTEM',
			'LEVEL'		: level,
			'MODULE'	: module,
			'MESSAGE'	: message,
		}
		self._write_db_log(log_data)

	def error(self, module, message):
		"""记录错误日志"""
		self._log('ERROR', module, message)

	def warn(self, module, message):
		"""记录警告日志"""
		self._log('WARN', module, message)

	def info(self, module, message):
		"""记录信息日志"""
		self._log('INFO', module, message)
	
	def debug(self, module, message):
		"""记录调试日志"""
		self._log('DEBUG', module, message)

log	= Logger(application_config['db'], application_config['path_log'])

if __name__ == '__main__':
	if os.getuid()!=0:
		print("ROOT only!")
		exit(1)
	if len(sys.argv)==1:
		# 无参数表示只启动后端（先停后启动）
		webframe_run_stop()
		webframe_run_start()
		# 打印配置参数
		for c in application_config:
			print(f"{c:>50}\\t{application_config[c]}")
	elif len(sys.argv)==2 and sys.argv[1] in ['stop']:
		# 只有这种情况停止
		webframe_run_stop()
	elif len(sys.argv)==2 and sys.argv[1] in ['start']:
		# 启动前端和后端（先停后启动）
		webframe_run_stop()
		webframe_run_start(backend_only=False)
		# 打印配置参数
		for c in application_config:
			print(f"{c:>50}\\t{application_config[c]}")
	else:
		print("参数错误")
		exit(1)
'''


#!/usr/bin/python3
# coding=utf-8
##################################################################
#              ____     _     _ __  __                 
#             / __/__  (_)___(_) /_/ /  ___  ___  ___ _
#            _\ \/ _ \/ / __/ / __/ /__/ _ \/ _ \/ _ `/
#           /___/ .__/_/_/ /_/\__/____/\___/_//_/\_, / 
#              /_/                              /___/  
# 
# Copyright (c) 2025 Chongqing Spiritlong Technology Co., Ltd.
# All rights reserved.
# @author	arthuryang
# @brief	SSH工具
##################################################################

# session名称格式：用户名@HOSTNAME。符合此格式的才进行处理

import	os
import	base64

# 仅在windows下可用
if os.name=='nt':
	import	winreg

import	hmac
import	hashlib
import	paramiko
import	SpiritLong_utility
import	SpiritLong_excel
from	nacl.signing	import	SigningKey	# pip install pynacl
from	nacl.encoding	import	Base64Encoder	as nacl_Base64Encoder
from	nacl.encoding	import	RawEncoder	as nacl_RawEncoder

# PuTTY注册表sessions路径
PUTTY_REGISTRY_PATH	= r"SOFTWARE\SimonTatham\PuTTY\Sessions"

# SSH路径：用户/OneDrive/文档/SSH
# 仅在windows下可用
if os.name=='nt':
	USER_SSH_BASE	= os.path.join(os.getenv('USERPROFILE'), 'OneDrive', '文档', 'SSH')
else:
	USER_SSH_BASE	= os.getenv('HOME')

# 需要读取/保存的设置
PUTTY_CONFIG	= [
	"HostName",		# 主机地址
	"PortNumber",		# 端口
	"UserName",		# 用户名
	"PublicKeyFile",	# 密钥文件（ppk文件）位置
]

# PuTTY注册表中的全部配置
PUTTY_CONFIG_DEFAULT	= {
	'HostName'			: '',
	'PortNumber'			: 22,
	'UserName'			: 'root',
	'PublicKeyFile'			: '',			# ppk文件位置
	
	# 字体
	'Font'				: 'Fixedsys',
	'FontHeight'			: 10,
	'FontIsBold'			: 0,			# 字体样式：是否加粗
	'FontQuality'			: 3,			# 字体质量 3->ClearType
	'FontCharSet'			: 134,			# 字体编码

	# Alt+Enter全屏
	'FullScreenOnAltEnter'		: 1,

	# 初始分辨率，单位为字符
	'TermWidth'			: 132,
	'TermHeight'			: 80,

	'WinTitle'			: 'arthuryang@PROGRAM',
	'Protocol'			: 'ssh',
	'BlinkCur'			: 1,
		
	# 行缓存数量	
	'ScrollbackLines'		: 999999,
	# X11转发使能才可以使用图形界面
	'X11Forward'			: 1,

	# TCP选项
	'PingInterval'			: 0,	# 发送TCPKeepalives的间隔：分钟数部分
	'PingIntervalSecs'		: 30,	# 发送TCPKeepalives的间隔：秒数部分
	'TCPNoDelay'			: 1,	# TCP_NODELAY
	'TCPKeepalives'			: 1,	# SO_KEEPALIVE

	'NoApplicationKeys'		: 1,	# 禁止应用程序小键盘，这可以在vi中正常使用小键盘
	'NoRemoteWinTitle'		: 1,	# 禁止远程修改标题

	# 以下是PuTTY默认设置
	'Present'			: 1,
	'LogFileName'			: 'putty.log',
	'LogType'			: 0,
	'LogFileClash'			: 4294967295,
	'LogFlush'			: 1,
	'LogHeader'			: 1,
	'SSHLogOmitPasswords'		: 1,
	'SSHLogOmitData'		: 0,
	'CloseOnExit'			: 1,
	'WarnOnClose'			: 1,	
	'TerminalType'			: 'xterm',
	'TerminalSpeed'			: '38400,38400',
	'TerminalModes'			: 'CS7=A,CS8=A,DISCARD=A,DSUSP=A,ECHO=A,ECHOCTL=A,ECHOE=A,ECHOK=A,ECHOKE=A,ECHONL=A,EOF=A,EOL=A,EOL2=A,ERASE=A,FLUSH=A,ICANON=A,ICRNL=A,IEXTEN=A,IGNCR=A,IGNPAR=A,IMAXBEL=A,INLCR=A,INPCK=A,INTR=A,ISIG=A,ISTRIP=A,IUCLC=A,IUTF8=A,IXANY=A,IXOFF=A,IXON=A,KILL=A,LNEXT=A,NOFLSH=A,OCRNL=A,OLCUC=A,ONLCR=A,ONLRET=A,ONOCR=A,OPOST=A,PARENB=A,PARMRK=A,PARODD=A,PENDIN=A,QUIT=A,REPRINT=A,START=A,STATUS=A,STOP=A,SUSP=A,SWTCH=A,TOSTOP=A,WERASE=A,XCASE=A',
	'AddressFamily'			: 0,
	'ProxyExcludeList'		: '',
	'ProxyDNS'			: 1,
	'ProxyLocalhost'		: 0,
	'ProxyMethod'			: 0,
	'ProxyHost'			: 'proxy',
	'ProxyPort'			: 80,
	'ProxyUsername'			: '',
	'ProxyPassword'			: '',
	'ProxyTelnetCommand'		: 'connect %host %port\n',
	'ProxyLogToTerm'		: 1,
	'Environment'			: '',
	'UserNameFromEnvironment	'	: 0,
	'LocalUserName'			: '',
	'NoPTY'				: 0,
	'Compression'			: 0,
	'TryAgent'			: 1,
	'AgentFwd'			: 0,
	'GssapiFwd'			: 0,
	'ChangeUsername'		: 0,
	'Cipher'			: 'aes,chacha20,3des,WARN,des,blowfish,arcfour',
	'KEX'				: 'ecdh,dh-gex-sha1,dh-group14-sha1,rsa,WARN,dh-group1-sha1',
	'HostKey'			: 'ed448,ed25519,ecdsa,rsa,dsa,WARN',
	'PreferKnownHostKeys'		: 1,
	'RekeyTime'			: 60,
	'GssapiRekey'			: 2,
	'RekeyBytes'			: '1G',
	'SshNoAuth'			: 0,
	'SshNoTrivialAuth'		: 0,
	'SshBanner'			: 1,
	'AuthTIS'			: 0,
	'AuthKI'			: 1,
	'AuthGSSAPI'			: 1,
	'AuthGSSAPIKEX'			: 1,
	'GSSLibs'			: 'gssapi32,sspi,custom',
	'GSSCustom'			: '',
	'SshNoShell'			: 0,
	'SshProt'			: 3,
	'LogHost'			: '',
	'SSH2DES'			: 0,
	'RemoteCommand'			: '',
	'RFCEnviron'			: 0,
	'PassiveTelnet'			: 0,
	'BackspaceIsDelete'		: 1,
	'RXVTHomeEnd'			: 0,
	'LinuxFunctionKeys'		: 0,
	'NoApplicationCursors'		: 0,
	'NoMouseReporting'		: 0,
	'NoRemoteResize'		: 0,
	'NoAltScreen'			: 0,
	'NoRemoteClearScroll'		: 0,
	'RemoteQTitleAction'		: 1,
	'NoDBackspace'			: 0,
	'NoRemoteCharset'		: 0,
	'ApplicationCursorKeys'		: 0,
	'ApplicationKeypad'		: 1,
	'NetHackKeypad'			: 0,
	'AltF4'				: 1,
	'AltSpace'			: 0,
	'AltOnly'			: 0,
	'ComposeKey'			: 0,
	'CtrlAltKeys'			: 1,
	'TelnetKey'			: 0,
	'TelnetRet'			: 1,
	'LocalEcho'			: 2,
	'LocalEdit'			: 2,
	'Answerback'			: 'PuTTY',
	'AlwaysOnTop'			: 0,
	'HideMousePtr'			: 0,
	'SunkenEdge'			: 0,
	'WindowBorder'			: 1,
	'CurType'			: 0,
	'Beep'				: 1,
	'BeepInd'			: 0,
	'BellWaveFile'			: '',
	'BellOverload'			: 1,
	'BellOverloadN'			: 5,
	'BellOverloadT'			: 2000,
	'BellOverloadS'			: 5000,
	'DECOriginMode'			: 0,
	'AutoWrapMode'			: 1,
	'LFImpliesCR'			: 0,
	'CRImpliesLF'			: 0,
	'DisableArabicShaping'		: 0,
	'DisableBidi'			: 0,
	'WinNameAlways'			: 1,
	'FontVTMode'			: 4,
	'UseSystemColours'		: 0,
	'TryPalette'			: 0,
	'ANSIColour'			: 1,
	'Xterm256Colour'		: 1,
	'TrueColour'			: 1,
	'BoldAsColour'			: 1,
	'Colour0'			: '187,187,187',
	'Colour1'			: '255,255,255',
	'Colour2'			: '0,0,0',
	'Colour3'			: '85,85,85',
	'Colour4'			: '0,0,0',
	'Colour5'			: '0,255,0',
	'Colour6'			: '0,0,0',
	'Colour7'			: '85,85,85',
	'Colour8'			: '187,0,0',
	'Colour9'			: '255,85,85',
	'Colour10'			: '0,187,0',
	'Colour11'			: '85,255,85',
	'Colour12'			: '187,187,0',
	'Colour13'			: '255,255,85',
	'Colour14'			: '0,0,187',
	'Colour15'			: '85,85,255',
	'Colour16'			: '187,0,187',
	'Colour17'			: '255,85,255',
	'Colour18'			: '0,187,187',
	'Colour19'			: '85,255,255',
	'Colour20'			: '187,187,187',
	'Colour21'			: '255,255,255',
	'RawCNP'			: 0,
	'UTF8linedraw'			: 0,
	'PasteRTF'			: 0,
	'MouseIsXterm'			: 0,
	'RectSelect'			: 0,
	'PasteControls'			: 0,
	'MouseOverride'			: 1,
	'Wordness0'			: '0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0',
	'Wordness32'			: '0,1,2,1,1,1,1,1,1,1,1,1,1,2,2,2,2,2,2,2,2,2,2,2,2,2,1,1,1,1,1,1',
	'Wordness64'			: '1,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,1,1,1,1,2',
	'Wordness96'			: '1,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,1,1,1,1,1',
	'Wordness128'			: '1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1',
	'Wordness160'			: '1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1',
	'Wordness192'			: '2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,1,2,2,2,2,2,2,2,2',
	'Wordness224'			: '2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,1,2,2,2,2,2,2,2,2',
	'MouseAutocopy'			: 1,
	'MousePaste'			: 'explicit',
	'CtrlShiftIns'			: 'explicit',
	'CtrlShiftCV'			: 'none',
	'LineCodePage'			: '',
	'CJKAmbigWide'			: 0,
	'UTF8Override'			: 1,
	'Printer'			: '',
	'CapsLockCyr'			: 0,
	'ScrollBar'			: 1,
	'ScrollBarFullScreen'		: 0,
	'ScrollOnKey'			: 0,
	'ScrollOnDisp'			: 1,
	'EraseToScrollback'		: 1,
	'LockSize'			: 0,
	'BCE'				: 1,
	'BlinkText'			: 0,
	'X11Display'			: '',
	'X11AuthType'			: 1,
	'X11AuthFile'			: '',
	'LocalPortAcceptAll'		: 0,
	'RemotePortAcceptAll'		: 0,
	'PortForwardings'		: '',
	'BugIgnore1'			: 0,
	'BugPlainPW1'			: 0,
	'BugRSA1'			: 0,
	'BugIgnore2'			: 0,
	'BugHMAC2'			: 0,
	'BugDeriveKey2'			: 0,
	'BugRSAPad2'			: 0,
	'BugPKSessID2'			: 0,
	'BugRekey2'			: 0,
	'BugMaxPkt2'			: 0,
	'BugOldGex2'			: 0,
	'BugWinadj'			: 0,
	'BugChanReq'			: 0,
	'StampUtmp'			: 1,
	'LoginShell'			: 1,
	'ScrollbarOnLeft'		: 0,
	'BoldFont'			: '',
	'BoldFontIsBold'		: 0,
	'BoldFontCharSet'		: 0,
	'BoldFontHeight'		: 0,
	'WideFont'			: '',
	'WideFontIsBold'		: 0,
	'WideFontCharSet'		: 0,
	'WideFontHeight'		: 0,
	'WideBoldFont'			: '',
	'WideBoldFontIsBold'		: 0,
	'WideBoldFontCharSet'		: 0,
	'WideBoldFontHeight'		: 0,
	'ShadowBold'			: 0,
	'ShadowBoldOffset'		: 1,
	'SerialLine'			: 'COM1',
	'SerialSpeed'			: 115200,
	'SerialDataBits'		: 8,
	'SerialStopHalfbits'		: 2,
	'SerialParity'			: 0,
	'SerialFlowControl'		: 1,
	'WindowClass'			: '',
	'ConnectionSharing'		: 0,
	'ConnectionSharingUpstream'	: 1,
	'ConnectionSharingDownstream'	: 1,
	'SSHManualHostKeys'		: '',
	'SUPDUPLocation'		: 'The Internet',
	'SUPDUPCharset'			: 0,
	'SUPDUPMoreProcessing'		: 0,
	'SUPDUPScrolling'		: 0,
}

## 在数据前面增加4字节长度，注意长度是大端
def big_Endian_bytes(data):
	if isinstance(data, str):
		# 字符串需要先转换到bytes
		data	= data.encode('ascii')
	return len(data).to_bytes(4, 'big')+data

## 生成随机密钥
def generate_random_key(algorithm='ssh-ed25519'):
	# 生成ed25519密钥
	key	= SigningKey.generate()
	verify	= key.verify_key

	# 生成密钥bytes
	private_key	= key.encode(nacl_RawEncoder)
	public_key	= verify.encode(nacl_RawEncoder)

	return {
		# 原始密钥
		'private'	: private_key,
		'public'	: public_key,

		# 密钥base64编码
		'private_string'	: base64.b64encode(big_Endian_bytes(private_key)).decode('ascii'),
		# 算法名称+密钥，然后base64编码
		'public_string'		: base64.b64encode(big_Endian_bytes(algorithm)+big_Endian_bytes(public_key)).decode('ascii'),
	}
	
## 从putty的ppk文件提取出密钥各部分信息
def _get_ppk_items(ppk_file_path):
	items	= {}
	with open(ppk_file_path, 'r', encoding='utf-8') as f:
		lines	= f.readlines()
		i	= 0
		while i<len(lines):
			n	= 1
			try:
				title, value	= lines[i].split(':', maxsplit=1)
				value	= value.strip()
				if title.startswith('PuTTY-User-Key-File'):
					# key的算法类型
					title	= 'key_type'
				if title=='Public-Lines':
					title	= 'public'
					n	= int(value)+1
					key_value_lines	= [v.strip() for v in lines[(i+1):(i+n)]]
					value	= ''.join(key_value_lines)
				elif title=='Private-Lines':
					title	= 'private'
					n	= int(value)+1
					key_value_lines	= [v.strip() for v in lines[(i+1):(i+n)]]
					value	= ''.join(key_value_lines)
				elif title=='Comment':
					title	= 'comment'
					value	= value.strip()
				items[title]	= value
			except:
				title	= None
				value	= None
			i	+= n;			
	return items


## 获取session在PUTTY_CONFIG中的配置
def get_session_config(session_name):
	# 仅在windows下可用
	if os.name!='nt':
		print("windows only")
		return

	config	= {}
	try:
		session_key		= winreg.OpenKey(winreg.HKEY_CURRENT_USER, f"{PUTTY_REGISTRY_PATH}\\{session_name}")
	
		for c in PUTTY_CONFIG:
			config[c], value_type	= winreg.QueryValueEx(session_key, c)
		winreg.CloseKey(session_key)
	except Exception as e:
		print(e)
	return config

## 从注册表读取PuTTY配置，写入到excel
def export_PuTTY_config(filename):
	# 仅在windows下可用
	if os.name!='nt':
		print("windows only")
		return

	sessions	= []
	try:
		# 读取putty的所有ses///sion的注册表
		session_root_key	= winreg.OpenKey(winreg.HKEY_CURRENT_USER, PUTTY_REGISTRY_PATH)
		subkeys_n, _, _	= winreg.QueryInfoKey(session_root_key)
		for i in range(subkeys_n):
			# 每个subkey是一个session，只取关键的配置
			session_name	= winreg.EnumKey(session_root_key, i)
			session		= get_session_config(session_name)

			# 校验session名的格式：user@HOSTNAME；若session名称里面包含了"."，则不是HOSTNAME，而是session名称，仍然放在HOSTNAME字段
			parts	= session_name.split('@')
			if len(parts)>1:
				# 包含@则必须是username@HOSTNAME的样式
				session['HOSTNAME']	= parts[1]
				if parts[0]!=session['UserName']:
					# 用户名得对得上才行
					session	= None
			elif '.' not in parts[0]:
				# 若不包含@，则应该是一个域名，必须包含"."，否则无效
				session	= None
			else:
				# 不包含@但包含"."的将session名称放在HOSTNAME
				session['HOSTNAME']	= parts[0]

			if session:
				sessions.append(session)
		winreg.CloseKey(session_root_key)
	except Exception as e:
		print(f"读取注册表失败: {e}")
	
	# 生成excel数据
	records	= []
	for s in sessions:
		r	= {
			'主机'		: s['HostName'],
			'HOSTNAME'	: s['HOSTNAME'],
			'端口'		: s['PortNumber'],
			'用户'	: s['UserName'],
		}
		
		generate	= True
		try:
			v	= _get_ppk_items(s['PublicKeyFile'])
			if v['key_type']=='ssh-ed25519':
				generate	= False
			r['私钥']	= v['private']
			r['公钥']	= v['public']
		except:
			pass

		if generate:
			# 生成密钥
			k	= generate_random_key()
			r['私钥']	= k['private_string']
			r['公钥']	= k['public_string']
		records.append(r)

	# 写入到excel
	SpiritLong_excel.records_to_excel(filename, {
		'主机'	: {
			# 确保按此顺序放置各列
			'titles'	: ['用户', 'HOSTNAME', '主机', '端口', '公钥', '私钥'],
			'records'	: records,
		}
	})

## 根据公钥和私钥生成ppk文件
#	filename	ppk文件名，含路径
#	public_key	公钥，带有长度的base64字符串（含算法名称）
#	private_key	私钥，带有长度的base64字符串
def generate_ppk_file(filename, public_key, private_key, comment):
	# 将字符串换成字节数组
	public_key	= public_key.encode('ascii')
	private_key	= private_key.encode('ascii')

	# 公钥分行
	public_lines	= [public_key[i:i+64].decode('ascii')+'\n' for i in range(0, len(public_key), 64)]

	# 私钥分行
	private_lines	= [private_key[i:i+64].decode('ascii')+'\n' for i in range(0, len(private_key), 64)]

	# 计算MAC，注意私钥和公钥要base64解码
	data	= big_Endian_bytes('ssh-ed25519')			+\
		  big_Endian_bytes('none')				+\
		  big_Endian_bytes(comment)				+\
		  big_Endian_bytes(base64.b64decode(public_key))	+\
		  big_Endian_bytes(base64.b64decode(private_key))
	MAC	= hmac.new(b'', data, hashlib.sha256).hexdigest()

	with open(filename, 'w', encoding='utf-8') as f:
		f.write("PuTTY-User-Key-File-3: ssh-ed25519\n")
		f.write("Encryption: none\n")
		f.write(f"Comment: {comment}\n")
		f.write(f"Public-Lines: {len(public_lines)}\n")
		f.writelines(public_lines)
		f.write(f"Private-Lines: {len(private_lines)}\n")
		f.writelines(private_lines)
		f.write(f"Private-MAC: {MAC}\n")

## 生成OPEN_SSH私钥文件（仅支持ed25519)
#	filename	ppk文件名，含路径
#	public_key	公钥，带有长度的base64字符串（含算法名称）
#	private_key	私钥，带有长度的base64字符串
def generate_OPENSSH_file(filename, public_key, private_key, comment=''): 
	# OpenSSH密钥格式 ed25519 private v1密钥格式，参考：https://sshref.dev/
	# 0.0 "openssh-key-v1" string plus terminating nullbyte (15 bytes)
	# 1.0 uint32 allocator for 1.0.0 (4 bytes)
	#     1.0.0 cipher name string (ASCII bytes)
	# 2.0 uint32 allocator for 2.0.0 (4 bytes)
	#     2.0.0 KDF name string (ASCII bytes)
	# 3.0 uint32 allocator for KDF options (3.0.0 to 3.0.1) (4 bytes) (ALWAYS 0 for unencrypted keys, so no following substructure)
	# 4.0 uint32 counter for # of keys (4 bytes)
	#     4.0.0 uint32 allocator for public key #n (4.0.0.0 to 4.0.0.1) (4 bytes)
	#         4.0.0.0 uint32 allocator for 4.0.0.0.0 (4 bytes)
	#             4.0.0.0.0 public key #n keytype string (ASCII bytes)
	#         4.0.0.1 uint32 allocator for 4.0.0.1.0 (4 bytes)
	#             4.0.0.1.0 public key #n payload (bytes)
	#     4.0.1 uint32 allocator for private key structure #n (4.0.1.0 to 4.0.1.5) (4 bytes)
	#         4.0.1.0 uint32 decryption "checksum" #1 (should match 4.0.1.1) (4 bytes)
	#         4.0.1.1 uint32 decryption "checksum" #2 (should match 4.0.1.0) (4 bytes)
	#         4.0.1.2 Copy of 4.0.0.0; allocator for 4.0.1.2.0 (4 bytes)
	#             4.0.1.2.0 Copy of 4.0.0.0.0 (ASCII bytes)
	#         4.0.1.3 Copy of 4.0.0.1; allocator for 4.0.1.3.0 (4 bytes)
	#             4.0.1.3.0 Copy of 4.0.0.1.0 (bytes)
	#         4.0.1.4 uint32 allocator for 4.0.1.4.0 (4 bytes)
	#             4.0.1.4.0 Private key #n (bytes)
	#         4.0.1.5 uint32 allocator for 4.0.1.5.0 (4 bytes)
	#             4.0.1.5.0 comment for key #n string (ASCII bytes)
	#         4.0.1.6 sequential padding

	key_n		= 1				# 密钥数量 (4.0 uint32 counter for # of keys)
	
	# 解算base64到bytes，注意公钥是类型+公钥
	public_bytes		= base64.b64decode(public_key)
	public_key_bytes	= public_bytes[(int.from_bytes(public_bytes[:4], 'big')+4+4):]

	# 解算base64到bytes
	private_bytes	= base64.b64decode(private_key)

	# 校验和1：未使用 (4.0.1.0 uint32 decryption "checksum" #1)
	# 校验和2：未使用 (4.0.1.1 uint32 decryption "checksum" #2)
	private_key_bytes	= big_Endian_bytes('')+big_Endian_bytes('')
	# 算法名（ssh-ed25519）(4.0.1.2.0 Copy of 4.0.0.0.0)
	# 公钥内容32字节 (4.0.1.3.0 Copy of 4.0.0.1.0)
	private_key_bytes	+=  public_bytes
	# 私钥内容64字节 (4.0.1.4.0 Private key #n (bytes))
	private_key_bytes	+=   big_Endian_bytes(private_bytes[4:]+public_key_bytes)
	# 私钥注释 (4.0.1.5.0 comment for key #n string)
	private_key_bytes	+=  big_Endian_bytes(comment)
	# sequential padding填充，私钥数据长度必须是8的倍数 (4.0.1.6 sequential padding)
	padding_length	= 8 - (len(private_key_bytes) % 8)
	if padding_length<8:
		private_key_bytes	+= bytes(range(1, 1+padding_length))	# 填充1,2,3...padding_length
	
	key_bytes	=  b'openssh-key-v1\0'			# OpenSSH密钥标识 (0.0 "openssh-key-v1" string plus terminating nullbyte)
	key_bytes	+= big_Endian_bytes('none')		# 加密算法：无 (1.0 uint32 allocator for 1.0.0 cipher name string)
	key_bytes	+= big_Endian_bytes('none')		# KDF算法：无 (2.0 uint32 allocator for 2.0.0 KDF name string)
	key_bytes	+= big_Endian_bytes('')			# KDF选项 (3.0 uint32 allocator for KDF options, ALWAYS 0 for unencrypted keys)
	key_bytes	+= key_n.to_bytes(4, 'big')		# 密钥数量
	key_bytes	+= big_Endian_bytes(public_bytes)	# 公钥
	key_bytes	+= big_Endian_bytes(private_key_bytes)	# 私钥

	# 编码成base64字符串
	key_string	= base64.b64encode(key_bytes).decode('ascii')

	# 写入文件
	with open(filename, 'w', encoding='utf-8') as f:
		f.write("-----BEGIN OPENSSH PRIVATE KEY-----\n")
		f.write('\n'.join(key_string[i:i+70] for i in range(0, len(key_string), 70)))
		f.write("\n-----END OPENSSH PRIVATE KEY-----\n")

## 根据配置文件写入PuTTY注册表选项
#	filename	配置文件名（含路径）
#	PPK_path	密钥保存路径，ppk文件将放在path/PPK下，OPENSSH文件将放在path/OPENSSH下
#	who_as_root	当用户为root时，需要个表示实际人员的字符串，此字符串将出现在/root/.ssh/authorized_keys里面该密钥的后面注释，以区分是谁的公钥
#	flush		清除当前所有配置，重新写入。注意如果此选项为True，则需要管理员权限
def import_PuTTY_config(filename, who_as_root, SSH_path=f'{USER_SSH_BASE}', flush=False, font_name='Fixedsys'):
	# 仅在windows下可用
	if os.name!='nt':
		print("windows only")
		return

	records	= []
	try:
		book	= SpiritLong_excel.open_xlsx(filename)
		records	= SpiritLong_excel.get_records_with_title(book['主机'])
	except:
		print(f"无法读取配置文件{filename}")
		return
	
	# 清除当前所有Session
	if flush:
		# 由于无法直接删除带子键的键，因此需要遍历
		session_root_key	= winreg.OpenKey(winreg.HKEY_CURRENT_USER, PUTTY_REGISTRY_PATH, access=winreg.KEY_ALL_ACCESS)
		subkeys_n, _, _	= winreg.QueryInfoKey(session_root_key)
		session_names	= []
		for i in range(subkeys_n):
			# Session键中应该没有子键
			session_names.append(winreg.EnumKey(session_root_key, i))
		for s in session_names:
			session_key	= winreg.OpenKey(winreg.HKEY_CURRENT_USER, f"{PUTTY_REGISTRY_PATH}\\{s}")
			winreg.DeleteKeyEx(session_key, '')
		winreg.CloseKey(session_root_key)
	
	# PPK文件所在路径
	PPK_path	= os.path.join(SSH_path, 'PPK')
	if not os.path.exists(PPK_path):
		os.mkdir(PPK_path)

	# OPENSSH文件所在路径
	OPENSSH_path	= os.path.join(SSH_path, 'OPENSSH')
	if not os.path.exists(OPENSSH_path):
		os.mkdir(OPENSSH_path)
 
	for r in records:
		# session名称：用户名@HOSTNAME，但若HOSTNAME有.则直接使用HOSTNAME作为session名称
		session_name	= r['HOSTNAME'] if '.' in r['HOSTNAME'] else f"{r['用户']}@{r['HOSTNAME']}"
		# 创建记录
		session_key	= winreg.CreateKey(winreg.HKEY_CURRENT_USER, f"{PUTTY_REGISTRY_PATH}\\{session_name}")
		
		comment		= f"{who_as_root}@{r['HOSTNAME']}" if r['用户']=='root' else f"{r['用户']}"
		
		# 生成ppk
		ppk_file	= os.path.join(PPK_path, session_name+'.ppk')
		generate_ppk_file(ppk_file, r['公钥'], r['私钥'], comment)

		# 生成OPENSSH
		OPENSSH_file	= os.path.join(OPENSSH_path, session_name+'.key')
		generate_OPENSSH_file(OPENSSH_file, r['公钥'], r['私钥'], comment)
		
		# 写入默认的配置值
		for k in PUTTY_CONFIG_DEFAULT:
			# 不是字符串就只可能是DWORD了
			winreg.SetValueEx(session_key, k, 0, winreg.REG_SZ if isinstance(PUTTY_CONFIG_DEFAULT[k], str) else winreg.REG_DWORD, PUTTY_CONFIG_DEFAULT[k])

		# 写入/修改每个session的特有配置值
		winreg.SetValueEx(session_key, 'HostName',	0,	winreg.REG_SZ,		str(r['主机'])	)	# 主机地址
		winreg.SetValueEx(session_key, 'PortNumber',	0,	winreg.REG_DWORD,	r['端口']	)	# 端口
		winreg.SetValueEx(session_key, 'UserName',	0,	winreg.REG_SZ,		r['用户']	)	# 用户名
		winreg.SetValueEx(session_key, 'PublicKeyFile',	0,	winreg.REG_SZ,		ppk_file	)	# PPK文件位置
		winreg.SetValueEx(session_key, 'WinTitle',	0,	winreg.REG_SZ,		session_name	)	# 窗口标题
		winreg.SetValueEx(session_key, 'Font',		0,	winreg.REG_SZ,		font_name	)	# 字体名称
		
		winreg.CloseKey(session_key)

## 连接到远程主机，返回连接对象
def _connect_session(session_name, base=USER_SSH_BASE):
	s		= get_session_config(session_name)
	OPENSSH_file	= os.path.join(base, 'OPENSSH', f'{session_name}.key')
	ssh		= paramiko.SSHClient()
	ssh.load_system_host_keys()
	ssh.set_missing_host_key_policy(paramiko.client.AutoAddPolicy())
	key	= paramiko.Ed25519Key.from_private_key_file(OPENSSH_file)
	ssh.connect(
		hostname	= s['HostName'],
		username	= s['UserName'],
		pkey		= key,
		port		= s['PortNumber'],
	)

	return ssh

## 使用SFTP上传文件，使用session
def upload_on_session(session_name, local_file, remote_file, base=USER_SSH_BASE):
	ssh	= _connect_session(session_name, base)
	sftp	= ssh.open_sftp()
	sftp.put(local_file, remote_file)
	sftp.close()
	ssh.close()

## 使用SFTP下载文件，使用session
def download_on_session(session_name, remote_file, local_file, base=USER_SSH_BASE):
	ssh	= _connect_session(session_name, base)
	sftp	= ssh.open_sftp()
	sftp.get(remote_file, local_file)
	sftp.close()
	ssh.close()

## 远程连接执行非交互命令，使用session，返回stdout+stderr
def execute_on_session(session_name, command, base=USER_SSH_BASE):
	ssh	= _connect_session(session_name, base)
	stdin, stdout, stderr	= ssh.exec_command(command, False)
	
	stdout_content	= stdout.read().decode(encoding='utf-8')
	stderr_content	= stderr.read().decode(encoding='utf-8')

	ssh.close()
	return stdout_content+stderr_content

if __name__ == '__main__':
	
	# 构建OneDrive文档SSH目录路径
	#me		= 'arthuryang'
	#config_xlsx	= 'a.xlsx'
	# 导入
	#import_PuTTY_config(config_xlsx, base, me, flush=True, font_name='Maple Mono NF CN')

	# 导出	
	# export_PuTTY_config(config_xlsx)

	#session_name	= 'arthuryang@PROGRAM'
	# result	= execute_on_session(session_name, 'cat .bashrc')
	# print(result)

	# download_on_session(session_name, '.bashrc', 'bashrc')
	# upload_on_session(session_name, 'bashrc', 'bashrc')

	pass


content	= '''#!/usr/bin/python3
# coding=utf-8
##################################################################
#               ____     _     _ __  __                 
#              / __/__  (_)___(_) /_/ /  ___  ___  ___ _
#             _\ \/ _ \/ / __/ / __/ /__/ _ \/ _ \/ _ `/
#            /___/ .__/_/_/ /_/\__/____/\___/_//_/\_, / 
#               /_/                              /___/  
# 
# Copyright (c) 2025 Chongqing Spiritlong Technology Co., Ltd.
# All rights reserved.
# @author	arthuryang
# @brief	webframe框架路由
##################################################################

# route和uwsgi入口需要分成两个文件来实现，因为application/route.py将会import全局实例routes
# 如果将routes放在webframe.py中，而webframe.py又引用了application/ruote.py，从而会导致循环引用卡死

import	datetime
import	time
import	json
import	secrets
import	string
import	crypt
import	jwt	# pip install pyjwt
import	requests
from	werkzeug	import Request
import	redis as redis_package
from	permission import get_login_user_permission

redis_client	= redis_package.Redis(connection_pool=redis_package.ConnectionPool(db=0, decode_responses=True))

## JsonEncoder处理datetime、Decimal
class BackendJsonEncoder(json.JSONEncoder):
	def default(self, obj):
		if isinstance(obj, datetime.datetime):
			return obj.strftime('%Y-%m-%d %H:%M:%S')
		elif isinstance(obj, datetime.date):
			return obj.strftime('%Y-%m-%d')
		elif isinstance(obj, decimal.Decimal):
			return str(obj).rstrip("0").rstrip(".")
		else:
			return json.JSONEncoder.default(self, obj)
## 路由类和实例
class Routes():
	def __init__(self):
		self.routes		= {}
		# 需要在外部初始化这些成员变量
		self.redis		= None
		self.db			= None
		
		# 默认TOKEN算法
		self.token_algorithm	= 'HS256'
		# token的超期天数
		self.token_expire_days	= 7
		# token的更新天数
		self.token_update_days	= 1
		# 默认token key
		self.token_secret_key	= ''.join(secrets.choice(string.ascii_letters + string.digits) for _ in range(20))

	## 装饰器，增加路由
	#       str     route_name	      路由名称
	#       list    parameters_required     参数列表
	#       bool    authorization_required  是否需要认证
	# decorator
	def add(self, route_name, parameters_required=[], authorization_required=True, **options):
		def decorator(f):
			# 增加路由规则
			self.routes[route_name] = {
				'function'			: f,				# 返回字节数组body
				'parameters_required'		: parameters_required,		# 必须要有的参数列表
				'authorization_required'	: authorization_required,       # 此路由是否需要认证才可访问
			}
			
			return f
		return decorator
	
	## 获取路由规则的内容
	def get(self, route_string):
		return self.routes.get(route_string)

	## 生成TOKEN
	def generate_token(self, user_ID):
		# 查询用户的JWT_KEY
		result	= self.db.query_redis('USER', [user_ID])
		if not result:
			# 没找到用户
			return ''
		
		# TOKEN所用的KEY如果在JWT_KEY字段，则可以通过更新该字段来使得该用户的所有TOKEN都失效
		secret_key	= result[0]['JWT_KEY'] if 'JWT_KEY' in result[0] else self.token_secret_key
		return jwt.encode(key=secret_key, algorithm=self.token_algorithm, headers={
				'alg'	: self.token_algorithm,
				'typ'	: 'JWT',
			},
			payload={
				'user_ID'		: user_ID,		# 用户ID，必须有
				'timestamp'		: time.time(),		# 签发时间戳
			})
	
	## 路由查找
	#	environment	从uWSGI传入的环境参数
	#	use_websocket	是否使用websocket
	# 返回data
	def query(self, environment, use_websocket=True):
		# 返回值data的默认值，全部键的定义都在此
		data    = {
			'request'	: Request(environment),
			'route'		: None,
			'route_name'	: environment.get('PATH_INFO'),
			'user_ID'	: 0,	# 0 表示未指定用户
			'message'	: '',	# （路由设置）响应详细消息，供前端参考
			'result'	: 'OK',	# （路由设置）响应结果字符串，供前端参考
			'return_data'	: {},	# （路由设置）返回数据：可以是bytes（直接发给前端）或对象（打包到JSON发给前端）
			'form'		: {},	# 来自表单的参数
			'files'		: {},	# 上传的文件
			'json'		: {},	# 来自JSON的参数 
			'args'		: {},	# 来自URL的参数
			'data'		: None,	# 请求中附带的原始数据
			'parameters'	: {},	# 处理后的实际传入参数，包括表单和JSON，JSON优先
			'token'		: '',	# 非空则为返回的TOKEN，供前端更新
			'response_code'	: '200',# http状态码
			'response_type'	: 'application/json',# 响应类型
		}

		# 查找路由
		data['route']		= self.get(data['route_name'])
		
		# 没找到路由
		if data['route'] is None:
			data['message'] = f"找不到路由：{data['route_name']}"
			data['result']  = 'ERROR'
			return data
		
		# 该路由无法执行
		if not callable(data['route']["function"]):
			data['message'] = f"路由[{data['route_name']}]函数无法执行"
			data['result']  = 'ERROR'
			return data
		
		# -------- 参数解析 -----------
		data['args']	= data['request'].args
		data['headers']	= data['request'].headers
		data['cookies']	= data['request'].cookies
		data['stream']	= data['request'].stream

		# 解析Content-Type
		content_type	= data['request'].headers.get('Content-Type')
		if content_type.startswith('application/x-www-form-urlencoded'):
			# 只有表单数据
			data['form']	= data['request'].form
		elif content_type.startswith('multipart/form-data'):
			# 表单数据和文件
			data['form']	= data['request'].form
			data['files']	= data['request'].files
		elif content_type.startswith('application/json'):
			# JSON
			data['json']	= data['request'].json

		# 获取websocket_code
		data['parameters']['websocket_code']	= data['request'].headers.get('Websocket-Code')
		# 总是获取附带的原始数据
		data['data']	= data['request'].get_data()

		# json和form合并到parameters，JSON的优先。注意form和json都是ImmutableMultiDict，是不可修改的
		data['parameters'].update(data['form'])
		data['parameters'].update(data['json'])

		# 检查接口所要求的参数
		parameter_missing       = []
		for p in data['route']['parameters_required']:
			if data['parameters'] and p not in data['parameters'].keys():
				parameter_missing.append(p)
		if parameter_missing:
			data['message'] = f"缺少必需的参数：{','.join(parameter_missing)}"
			data['result']  = 'PARAMETER MISSING'
			# 缺少参数
			data['response_code']	= '400'
			return data

		# ------------- 认证处理 -------------
		# 认证的本质是获得user_ID
		if data['route']['authorization_required']:
			# 此路由需要认证
			authorization   = data['request'].headers.get('Authorization')
			
			if use_websocket:
				if 'websocket_code' not in data['parameters']:
					# 需要认证的必须上传websock_code
					data['message'] = f"认证必须要有参数websocket_code"
					data['result']  = 'ERROR'
					# 认证失败
					data['response_code']	= '401'
					return data
				websocket_code  = data['parameters']['websocket_code']

			if not authorization:
				# 不包含Authorization域
				data['message'] = f"认证Header未包含Authorziation"
				data['result']  = 'ERROR'
				# 认证失败
				data['response_code']	= '401'
				return data
			
			authorization   = authorization.split()
			if authorization[0] != 'Bearer' or len(authorization)==1:
				# 必须是Bearer类型的认证，且包含内容
				data['message'] = f"认证Authorziation未包含Bearer"
				data['result']  = 'ERROR'
				# 认证失败
				data['response_code']	= '401'
				return data
			
			# 解析TOKEN，其中必须包含user_ID
			token   = authorization[1] if len(authorization)>1 else None
			try:
				payload = jwt.decode(token, self.token_secret_key, [self.token_algorithm])
				# 解析TOKEN
				data['user_ID']		= payload['user_ID']
								
				# 判断是否超期失效
				expire_seconds	= time.time()-payload['timestamp']
				if expire_seconds>self.token_expire_days*86400:
					# 已经超期失效了
					data['message'] = f"认证TOKEN已失效"
					data['result']  = 'TOKEN EXPIRE'
					# 认证失败
					data['response_code']	= '401'
					return data
				elif expire_seconds>self.token_update_days:
					# 定期更换token
					data['token']	= self.generate_token(data['user_ID'])

			# 解析失败：TOKEN签名错误
			except jwt.ExpiredSignatureError:
				data['message'] 	= f"认证TOKEN签名错误"
				data['result']  	= 'TOKEN ERROR'
				data['response_code']	= '401'
				return data
			
			# 解析失败：非法TOKEN
			except jwt.InvalidTokenError:
				data['message'] 	= f"认证TOKEN非法"
				data['result']  	= 'TOKEN INVALID'
				# 认证失败
				data['response_code']	= '401'
				return data

			# user_ID绑定到websocket_code，redis中WEBSOCKET USER是一个hash类型，每个键是一个websocket_code，对应值为user_ID，0表示未绑定
			if use_websocket:
				# 更新
				self.redis.hset('WEBSOCKET USER', websocket_code, data['user_ID'])
		return data

	## 执行路由，返回响应
	#	data	来自query的返回值
	def response(self, data, start_response):
		# 执行路由, data中的result/message/return_data将会被修改
		print(f"==============response_code: {data['response_code']}")
		if data['route'] and data['response_code']=='200':
			data['route']["function"](data)

		# return_data是返回的数据
		if isinstance(data['return_data'], bytes):
			body	= data['return_data']
		elif data['response_type'] == 'text/html':
			body	= data['return_data'].encode('utf8')
		else:
			return_data	= {
				'result'	: data['result'],
				'message'       : data['message'],
				'data'		: data['return_data'],
			}
			return_data['token']	= data['token']
			body    = bytes(json.dumps(return_data, cls=BackendJsonEncoder), encoding='utf8')

		headers	 = []
		# 增加头部中的body长度字节数
		headers.append(('Content-Length', str(len(body))))
		
		# 使用路由指定的内容类型
		headers.append(('Content-Type', data['response_type']))
		
		# 总是返回状态200
		start_response(data['response_code'], headers)
		return body

# 全局的路由实例
routes = Routes()

# ###################################### 默认路由 ########################################

## 使用密码登录，用户名尝试依次在NAME、IDENTITY、MOBILE、EMAIL、USER里面搜索
@routes.add('/login_password', authorization_required=False, parameters_required=[
	'username',	# 用户名
	'password',	# 密码
])
def API_login_password(route_data):
	global routes

	parameters	= route_data['parameters']

	# 先假定登录失败
	route_data['message']	= "登录失败"
	route_data['result']	= "LOGIN_FAILED"

	# 搜索数据库表获得user_ID
	for field in ['NAME', 'IDENTITY', 'MOBILE', 'EMAIL', 'USER']:
		SQL_string	= f\'\'\'
			SELECT ID, PASSWORD
			FROM USER
			WHERE {field}='{parameters['username']}'
			AND DELETE_TIMESTAMP IS NULL
		\'\'\'
		result	= routes.db.query(SQL_string)
		if result:
			break
	if not result:
		# 没找到结果
		return
	
	salt_shadow	= result[0]['PASSWORD']
	salt		= salt_shadow[:12]
	shadow		= salt_shadow[12:]
	crypted		= crypt.crypt(parameters['password'], salt)
	if crypted==salt_shadow:
		# 密码验证通过
		user_ID	= result[0]['ID']
		route_data['token']	= routes.generate_token(user_ID)
		route_data['message']	= '登录成功'

## 使用微信登录
@routes.add('/login_wechat', authorization_required=False, parameters_required=[
	'wechat_code',		# 微信code
	'websocket_code',	# websocket code
])
def API_login_wechat(route_data):
	global routes

	parameters	= route_data['parameters']
	wechat_code	= parameters['wechat_code']
	websocket_code	= parameters['websocket_code']
	app_ID		= 'wx439a3874a5b28c4f'
	app_secret	= '613bcdf2937396290a9a550aedfcb2d7'

	url		= f\'\'\'https://api.weixin.qq.com/sns/oauth2/access_token?appid={app_ID}&secret={app_secret}&code={wechat_code}&grant_type=authorization_code\'\'\'
	response	= requests.get(url)
	if 'openid' not in response.json():
		print(f'获取openid失败: {response.json()}')
		route_data['response_code']	= '401'
		route_data['result']		= 'LOGIN_FAILED'
		route_data['message']		= '获取openid失败'
		return
	open_ID	= response.json()['openid']
	# 是否已有该用户的open_ID
	exist_user	= routes.db.query(f\'\'\'
		SELECT
			*
		FROM USER
		WHERE
			OPENID='{open_ID}'
		AND
			DELETE_TIMESTAMP IS NULL
	\'\'\')
	if not exist_user:
		# 创建用户信息
		exist_user	= routes.db.insert('USER', {
			'OPENID'	: open_ID,
			'NAME'		: '无',
			'IDENTITY'	: '无',
			'REGION'	: '无',
			'MOBILE'	: '无',
			'EMAIL'		: '无',
			'TYPE'		: '无',
		})
	user_information	= exist_user[0]

	# 获取用户组和权限
	get_login_user_permission(user_information['ID'])

	# 生成token
	timestamp				= time.time()
	token					= jwt.encode({
		'timestamp'		: timestamp,
		'user_ID'		: user_information['ID']
	}, routes.token_secret_key, algorithm=routes.token_algorithm)
	# 推送token到客户端
	message	= f"TOKEN {websocket_code} {token}"
	redis_client.publish('WEBSOCKET SERVER', message)
	
	# 推送用户信息到客户端
	data	= json.dumps({
		'ID'		: user_information['ID'],
		'NAME'		: user_information['NAME'],
	}, ensure_ascii=False)
	message	= f"USER {websocket_code} {data}"
	redis_client.publish('WEBSOCKET SERVER', message)

	route_data['data']	= {}
	route_data['message']	= '登录成功'
	route_data['result']	= 'LOGIN_SUCCESS'
	
## 微信扫码获取wechat_code
@routes.add('/authorize_redirect', authorization_required=False)
def API_authorize_redirect(route_data):
	redirect_URL	= route_data['args'].get('redirect_URL', '')
	app_ID		= route_data['args'].get('app_ID', '')

	# 读取template/authorize_redirect.html
	with open('template/authorize_redirect.html', 'r', encoding='utf-8') as f:
		html_content	= f.read()
		html_content	= html_content.replace('$$redirect_URL$$', redirect_URL)
		html_content	= html_content.replace('$$app_ID$$', app_ID)
		route_data['return_data']	= html_content
	print(f'redirect_URL: {redirect_URL}')
	print(f'app_ID: {app_ID}')

	route_data['response_type']	= 'text/html'
	
## 微信扫码返回授权页面
@routes.add('/authorize_scan', authorization_required=False)
def API_authorize_scan(route_data):
	websocket_code	= route_data['args'].get('websocket_code', '')
	app_ID		= route_data['args'].get('app_ID', '')

	# 读取template/authorize_scan_page.html
	with open('template/authorize_scan_page.html', 'r', encoding='utf-8') as f:
		html_content	= f.read()
		html_content	= html_content.replace('$$websocket_code$$', websocket_code)
		html_content	= html_content.replace('$$app_ID$$', app_ID)
		route_data['return_data']	= html_content
	route_data['response_type']	= 'text/html'
'''
